///////////////////////////////////////////////////////////////////////////////
//                                                         
// BigDecimal.cc
// -------------
// BigDecimal implementation
//                                               
// Design and Implementation by Bjoern Lemke               
//                                                         
// (C)opyright 2000-2007 Bjoern Lemke                        
//                                                         
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: BigDecimal
// 
// Description: BigDecimal implementation
//
// Status: IMPLEMENTED
//
///////////////////////////////////////////////////////////////////////////////


// INCLUDES
#include "Exception.h"
#include "BigInteger.h"
#include "BigDecimal.h"
#include "Tokenizer.h"

BigDecimal::BigDecimal()
{
    _val = Chain(0);
    _dim=0;
    _scale=0;
    _isPositive = true;
}

BigDecimal::BigDecimal(int dim, int scale)
{
    _val = Chain(0);
    _dim=dim;
    _scale=scale;
    _isPositive = true;
}

BigDecimal::BigDecimal(const Chain& val)
{
    if ( val.subChain(1,1) == Chain("+"))
    {
	_isPositive=true;
	_val = val.subChain(2, val.length());
    }
    else if ( val.subChain(1,1) == Chain("-"))
    {
	_isPositive=false;
	_val = val.subChain(2, val.length());
    }
    else
    {
	_isPositive=true;
	_val = val;
    }

    if ( ! _val.isDec() )
	throw Exception(EXLOC, "Invalid decimal format");

    Tokenizer tok(_val, Chain("."));
    Chain left, right;
    tok.nextToken(left);
    tok.nextToken(right);
    
    _scale = right.length() - 1;
    _val = left + right;
}

BigDecimal::BigDecimal(const Chain& val, int scale)
{

    if ( val.subChain(1,1) == Chain("+"))
    {
	_isPositive=true;
	_val = val.subChain(2, val.length());
    }
    else if ( val.subChain(1,1) == Chain("-"))
    {
	_isPositive=false;
	_val = val.subChain(2, val.length());
    }
    else
    {
	_isPositive=true;
	_val = val;
    }

    if ( ! _val.isDec() )
	throw Exception(EXLOC, "Invalid decimal format");

    _scale = scale;
}


BigDecimal::~BigDecimal()
{
}

BigDecimal BigDecimal::abs() const
{
    return BigDecimal(_val);
}

bool BigDecimal::isPositive() const
{
    return _isPositive;
}

void BigDecimal::negate()
{
    _isPositive = false;
}


BigDecimal BigDecimal::add(const BigDecimal& d) const
{
    Chain v1 = _val;
    Chain v2 = d._val;
    int maxScale;


    if ( _scale > d._scale )
    {
	maxScale = _scale;
	v2 = mulDec(v2, _scale - d._scale);
    }
    else if ( _scale < d._scale )
    {
	maxScale = d._scale;
	v1 = mulDec(v1, d._scale - _scale);
    }    
    else
    {
	maxScale = _scale;
    }

    BigInteger i1(v1);
    BigInteger i2(v2);

    if ( _isPositive == false )
	i1.negate();
    if ( d._isPositive == false )
	i2.negate();


    BigInteger i3 = i1 + i2;

    return BigDecimal(i3.toChain(), maxScale); 
    
}

BigDecimal BigDecimal::sub(const BigDecimal& d) const
{

    Chain v1 = _val;
    Chain v2 = d._val;
    int maxScale;

    if ( _scale > d._scale )
    {
	maxScale = _scale;
	v2 = mulDec(v2, _scale - d._scale);
    }
    else if ( _scale < d._scale )
    {
	maxScale = d._scale;
	v1 = mulDec(v1, d._scale - _scale);
    }    
    else
    {
	maxScale = _scale;
    }
    BigInteger i1(v1);
    if ( _isPositive == false )
	i1.negate();
    BigInteger i2(v2);
    if ( d._isPositive == false )
	i2.negate();


    BigInteger i3 =  i1 - i2;

    return BigDecimal(i3.toChain(), maxScale); 

}

BigDecimal BigDecimal::mul(const BigDecimal& d) const
{

    Chain v1 = _val;
    Chain v2 = d._val;

    BigInteger i1(v1);
    if ( _isPositive == false )
	i1.negate();
    BigInteger i2(v2);
    if ( d._isPositive == false )
	i2.negate();
    
    BigInteger i3 =  i1 * i2;

    return BigDecimal(i3.toChain(), _scale + d._scale); 
    
}

BigDecimal BigDecimal::div(const BigDecimal& d) const
{

    Chain v1 = _val;
    Chain v2 = d._val;

    int maxScale;
    
    if ( d._scale > _scale )
    {	
	maxScale = d._scale;
	v1 = mulDec(v1, 2 * d._scale - _scale);
    }
    else
    {
	maxScale = _scale;
	v2 = mulDec(v2, _scale - d._scale);
    }

    BigInteger i1(v1);
    if ( _isPositive == false )
	i1.negate();
    BigInteger i2(v2);
    if ( d._isPositive == false )
	i2.negate();
    
    BigInteger i3 =  i1 / i2;

    Chain v = i3.toChain();
    if ( i3.toChain().length()-1 < maxScale )
    {
	int c = 1 + maxScale - ( i3.toChain().length() -1 );
	while ( c )
	{
	    v = Chain("0") + v;
	    c--;
	}
    }
    return BigDecimal(v, maxScale); 

}

int BigDecimal::getScale() const
{
    return _scale;
}

int BigDecimal::length() const
{
    return _val.length();
}

Chain BigDecimal::toChain() const
{
    Chain s;

    if ( _isPositive == false )
	s = Chain("-");

    Chain val = _val;
    
    if ( val.length() - 1  <= _scale )
    {
	int j = val.length() - 1;
	for ( int i = j; i <= _scale ; i++)
	{
	    val = "0" + val;
	}
    }
    
    s +=  val.subChain(1, val.length() - _scale - 1) 
	+ Chain(".")
	+ val.subChain(val.length() - _scale, val.length()); 	

    return s;
}

BigDecimal& BigDecimal::operator = ( const BigDecimal& d)
{
    _val = d._val;
    _isPositive = d._isPositive;
    _scale = d._scale;
    _dim = d._dim;
    return (*this);
}

BigDecimal& BigDecimal::operator += ( const BigDecimal& d)
{
    this->add(d);
    return (*this);
}

BigDecimal& BigDecimal::operator -= ( const BigDecimal& d)
{
    this->add(d);
    return (*this);
}

bool BigDecimal::operator == ( const BigDecimal& d) const
{
    return ( _val == d._val && _isPositive == d._isPositive && _scale == d._scale);
}

bool BigDecimal::operator != ( const BigDecimal& d) const
{
    return ( _val != d._val || _isPositive != d._isPositive || _scale != d._scale );
}

bool BigDecimal::operator < ( const BigDecimal& d) const
{
    if ( ( _val.length() - _scale  ) < ( d._val.length() - d._scale ) )
	return true;
    else if ( ( _val.length() - _scale  ) > ( d._val.length() - d._scale ) )
	return false;
    else
    {
	if ( _scale < d._scale )
	{
	    BigInteger v1( d._val.subChain(1, d._val.length() - ( d._scale - _scale ) - 1));
	    BigInteger v2( _val );
	    return v1 < v2;
	}
	else
	{
	    BigInteger v1( d._val );
	    BigInteger v2( _val.subChain(1, _val.length() - ( _scale - d._scale ) - 1));
	    return v1 < v2;
	}
    }		
}

bool BigDecimal::operator > ( const BigDecimal& d) const
{
    if ( ( _val.length() - _scale  ) > ( d._val.length() - d._scale ) )
	return true;
    else if ( ( _val.length() - _scale  ) < ( d._val.length() - d._scale ) )
	return false;
    else
    {
	if ( _scale < d._scale )
	{
	    BigInteger v1( d._val.subChain(1, d._val.length() - ( d._scale - _scale ) - 1));
	    BigInteger v2( _val );
	    return v1 > v2;
	}
	else
	{
	    BigInteger v1( d._val );
	    BigInteger v2( _val.subChain(1, _val.length() - ( _scale - d._scale ) - 1));
	    return v1 > v2;
	}
    }
}

bool BigDecimal::operator <= ( const BigDecimal& d) const
{
    if ( _val == d._val && _isPositive == d._isPositive)
	return true;
    else
	return *this < d;
	
}

bool BigDecimal::operator >= ( const BigDecimal& d) const
{
    if ( _val == d._val && _isPositive == d._isPositive)
	return true;
    else
	return *this > d;

}

BigDecimal operator + ( const BigDecimal& d1, const BigDecimal& d2)
{
    return d1.add(d2);
}

BigDecimal operator - ( const BigDecimal& d1, const BigDecimal& d2)
{
    return d1.sub(d2);
}

BigDecimal operator * ( const BigDecimal& d1, const BigDecimal& d2)
{
    return d1.mul(d2);
}

BigDecimal operator / ( const BigDecimal& d1, const BigDecimal& d2)
{
    return d1.div(d2);
}

ostream& operator << (ostream& s, const  BigDecimal& d)
{
    cout << d.toChain();
    return s;
}

Chain BigDecimal::mulDec(const Chain& val, int dim) const
{
    Chain res = val;
    while ( dim > 0 )
    {
	res += Chain("0");
	dim--;
    }
    return res;
}

