///////////////////////////////////////////////////////////////////////////////
//                                                         
// NetHandler.cc
// -------------
// Net handle for sending and receiving messages
//                                               
// Design and Implementation by Bjoern Lemke               
//                                                         
// (C)opyright 2000-2007 Bjoern Lemke                        
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: NetHandler
// 
// Description: Net handle for sending and receiving messages
//
// Status: CLEAN
//
///////////////////////////////////////////////////////////////////////////////


#ifndef _REENTRANT
#define _REENTRANT    /* basic 3-lines for threads */
#endif

#include <unistd.h>
#include <string.h>
#include <sys/types.h>

#ifdef HAVE_MINGW32
#include <windows.h>
#include <winsock2.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#endif

#include <errno.h>

#include "Exception.h"
#include "Net.h"
#include "NetHandler.h"

#define SENDLEN 1024

NetHandler::NetHandler(int msgBufSize, int sizeInfoLen)
{
    _socket=0;
    _msgSize=0;
    _msgBufSize = msgBufSize;
    _sizeInfoLen = sizeInfoLen;

    _msgBuf = new char[_msgBufSize + _sizeInfoLen];
}

NetHandler::NetHandler(int s, int msgBufSize, int sizeInfoLen)
{
    _socket=s;
    _msgSize=0;
    _msgBufSize = msgBufSize;
    _sizeInfoLen = sizeInfoLen;

    _msgBuf = new char[_msgBufSize + _sizeInfoLen];
}

NetHandler::~NetHandler()
{
    if ( _socket )
	::close(_socket);
    delete _msgBuf;
}

int NetHandler::getSocket()
{
    return _socket;
}

/* obsolete, since the socket option is no supported by all platforms

  
void NetHandler::setSendTimeout(int timeout)
{

#ifdef HAVE_MINGW32

#else
    struct timeval tv;
    tv.tv_sec = timeout;
    tv.tv_usec = 0;
    
    if (::setsockopt(_socket, SOL_SOCKET, SO_SNDTIMEO, &tv, sizeof( struct timeval)) != 0 )
    {
	// Chain msg = Chain("setsockopt system error : ") + Chain(strerror(errno));
	// throw Exception(EXLOC, msg);

	// since the send timeout property is not supported on all 
	// platforms ( e.g. Solaris ) , we ignore this error
    }
#endif
}


void NetHandler::setRecvTimeout(int ms)
{
    struct timeval tv;
    
    tv.tv_sec = ms / 1000 ;
    tv.tv_usec = ( ms % 1000) * 1000  ;
    
    if  ( setsockopt (_socket, SOL_SOCKET, SO_RCVTIMEO, (char*)&tv, sizeof(struct timeval)) != 0 )
    {
	// since the send timeout property is not supported on all 
	// platforms ( e.g. Solaris ) , we ignore this error

	// Chain msg = Chain("setsockopt system error : ") + Chain(strerror(errno));
	// throw Exception(EXLOC, msg);
	
    }
}

*/

bool NetHandler::waitMsg(int ms)
{
    int result;

    struct timeval tv;
    
    tv.tv_sec = ms / 1000 ;
    tv.tv_usec = ( ms % 1000) * 1000  ;
        
    fd_set readset;
    
    do {
	FD_ZERO(&readset);
	FD_SET(_socket, &readset);
	result = select(_socket + 1, &readset, NULL, NULL, &tv);
    } while (result == -1 && errno == EINTR);
    
    if (result > 0) 
    {
	if (FD_ISSET(_socket, &readset)) 
	{
	    return true;
	}
    }
    else if (result < 0) 
    {
	Chain msg = Chain("select system error : ") + Chain(strerror(errno));
	throw Exception(EXLOC, msg);	
    }
    
    return false;
}
    
void NetHandler::readMsg()
{
    
    int recvByte;
    if ( (recvByte = ::recv(_socket,_msgBuf, _msgBufSize + _sizeInfoLen , 0 )) <= 0 )
    {
	Chain msg = Chain("recv system error : ") + Chain(strerror(errno));
	throw Exception(EXLOC, msg);
    }

    int i = 0;
    while ( _msgBuf[i] != '@' && i < _sizeInfoLen)
    {
	i++;
    }
    Chain sizeInfo(_msgBuf, i);
    _msgSize = sizeInfo.asInteger();
    
    if ( _msgSize > _msgBufSize )
    {
	char* copyBuf = _msgBuf;
	_msgBufSize = _msgSize + 1;
	_msgBuf = new char[_msgBufSize + _sizeInfoLen];
	memcpy(_msgBuf, copyBuf, recvByte );
	delete copyBuf;

    }
    while ( recvByte < _msgSize + _sizeInfoLen )
    {
	int i;
	if ( (i = ::recv(_socket, (char*)((long)_msgBuf + (long)recvByte), SENDLEN, 0)) <= 0 )
	{
	    Chain msg = Chain("recv system error : ") + Chain(strerror(errno));
	    throw Exception(EXLOC, msg);
	}
	recvByte += i;
    }
	
    // terminate msg in every case
    if ( _msgSize < _msgBufSize )
	_msgBuf[_msgSize + _sizeInfoLen]=0;

    return;
}

void NetHandler::writeMsg()
{
    // first send msg size

    Chain sizeInfo(_msgSize);
   
    if ( sizeInfo.length() > _sizeInfoLen )
    {
	Chain msg = Chain("Message too long : ") + sizeInfo;
	throw Exception(EXLOC, msg);
    }
    
    memcpy(_msgBuf, (char*)sizeInfo, sizeInfo.length() - 1 );
    
    int i = sizeInfo.length() - 1;
    while ( i < _sizeInfoLen)
    {
	_msgBuf[i] = '@';
	i++;
    }
    
    int sendByte = 0;
    while ( sendByte < _msgSize + _sizeInfoLen)
    {
	int i;
	
	int sendSize;

	if ( SENDLEN < _msgSize + _sizeInfoLen - sendByte )
	    sendSize =  SENDLEN; 
	else 
	    sendSize = _msgSize + _sizeInfoLen - sendByte;

	if ( ( i = ::send(_socket, (char*)((long)_msgBuf + (long)sendByte), sendSize,0)) == -1)
	{
	    Chain msg = Chain("send system error : ") + Chain(strerror(errno));
	 
	    throw Exception(EXLOC, msg);
	}
	
	sendByte += i;
		
    }
}

bool NetHandler::recvAck()
{
    
    char ack;
    int ackSize;
    if ( (ackSize = ::recv(_socket,&ack,1,0)) <= 0 )
    {
	Chain msg = Chain("recv system error : ") + Chain(strerror(errno));
	throw Exception(EXLOC, msg);
    }
    if ( ack == 1 )
    {
	return true;
    }
    else
    {
	return false;
    }
}

void NetHandler::sendAck()
{
    char ack = 1;
    int ackSize;
    
    if ( ( ackSize = ::send(_socket, &ack,1,0)) == -1)
    {
	Chain msg = Chain("send system error : ") + Chain(strerror(errno));
	throw Exception(EXLOC, msg);
    }
}

void NetHandler::sendNack()
{
    char ack = 0;
    int ackSize;
    
    if ( ( ackSize = ::send(_socket, &ack,1,0)) == -1)
    {
	Chain msg = Chain("send system error : ") + Chain(strerror(errno));
	throw Exception(EXLOC, msg);
    }
}

void NetHandler::sendChar(char c)
{
    
    if ( ::send(_socket, &c,1,0) == -1)
    {
	Chain msg = Chain("send system error : ") + Chain(strerror(errno));
	throw Exception(EXLOC, msg);
    }
}

char NetHandler::recvChar()
{
    
    char c;
    if ( ::recv(_socket,&c,1,0) <= 0 )
    {
	Chain msg = Chain("recv system error : ") + Chain(strerror(errno));
	throw Exception(EXLOC, msg);
    }
    return c;
}


char* NetHandler::getMsg()
{
    return (char*)( (long)_msgBuf + _sizeInfoLen);
}

int NetHandler::getMsgSize()
{
    return _msgSize;
}


void NetHandler::setMsg(char* msg, int size)
{
    if ( size > _msgBufSize )
    {
	delete _msgBuf;
	_msgBufSize = size + 1;
	_msgBuf = new char[_msgBufSize + _sizeInfoLen];
	// Chain msg = Chain("Message too long : ") + Chain(size) + Chain(" bytes");
	// throw Exception(EXLOC, msg);
    }
    if ( size > 0 )
	memcpy((char*)((long)_msgBuf + _sizeInfoLen), msg, size);

    _msgSize=size;
}


void NetHandler::disconnect()
{
    if (_socket )
    {
	// cout << "Closing socket " << _socket << endl;
	::close(_socket);
	_socket=0;
    }
}

NetHandler& NetHandler::operator=(const NetHandler& nh)
{
    _socket = nh._socket;
    memcpy(_msgBuf, nh._msgBuf, _msgBufSize + _sizeInfoLen);
    _msgSize = nh._msgSize;
    _msgBufSize = nh._msgBufSize;
    
    return (*this);
}
    
bool NetHandler::operator==(const NetHandler& nh) const
{
    if (nh._socket == _socket) 
	return true;
    return false;
}
