///////////////////////////////////////////////////////////////////////////////
//
// ThreadLock.cc 
// -------------
// ThreadLock class implementation
//                                                
// Design and Implementation by Bjoern Lemke               
//                                                         
// (C)opyright 2000-2007 Bjoern Lemke                        
//                                                         
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
// 
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; see the file COPYING.  If not, write to
// the Free Software Foundation, 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.
//
// IMPLEMENTATION MODULE
//
// Class: Thread
//
// Description: all operations on POSIX threads
//
// Status: IMPLEMENTED
///////////////////////////////////////////////////////////////////////////////

#ifndef _REENTRANT
#define _REENTRANT    /* basic 3-lines for threads */
#endif

#ifdef HAVE_MINGW32
#else
#include <pthread.h>
#endif
#include <errno.h>
#include <string.h>
#include <time.h>
#include <iostream>

#include "ThreadLock.h"
#include "Exception.h"
#include "NanoTimer.h"

#define DELAYFAC 1000000

ThreadLock::ThreadLock(const Chain& id)
{
    _id = id;
}

ThreadLock::~ThreadLock()
{
}

const Chain& ThreadLock::getId() const
{
    return _id;
}

void ThreadLock::setId(const Chain& id)
{
    _id = id;
}

void ThreadLock::init(int lockDelay, bool doStat)
{


#ifdef HAVE_MINGW32
    InitializeCriticalSection (& _rwlock);
#else

    _doStat = doStat;
    _readLockCount=0;
    _writeLockCount=0;
    _numLockTry=0;

    _readDelay = 0;
    _writeDelay = 0;

    _lockDelay = lockDelay;

    if ( pthread_rwlock_init(&_rwlock, NULL) != 0 )
    {
	Chain msg = Chain("ThreadLock system error : ") + Chain(strerror(errno));
	throw Exception(EXLOC, msg);
    }

    if ( _doStat )
    {
	if ( pthread_rwlock_init(&_statlock, NULL) != 0 )
	{
	    Chain msg = Chain("ThreadLock system error : ") + Chain(strerror(errno));
	    throw Exception(EXLOC, msg);
	}
    }
#endif	    
}

void ThreadLock::readLock()
{   

#ifdef HAVE_MINGW32
    EnterCriticalSection (& _rwlock);
#else

    NanoTimer *pTimer = 0;

    if ( _doStat )
    {
	pTimer = new NanoTimer();
	pTimer->start();

	pthread_rwlock_wrlock(&_statlock);
	_numLockTry++;
	_readLockCount++;
	pthread_rwlock_unlock(&_statlock);

    }

    int ret = pthread_rwlock_rdlock(&_rwlock);

    if ( _doStat )
    {
	pTimer->stop();

	pthread_rwlock_wrlock(&_statlock);
	_numLockTry--;
	_readDelay += pTimer->getSum();
	pthread_rwlock_unlock(&_statlock);

	delete pTimer;
	
    }
    
    if ( ret != 0 )
    {
	
	Chain msg = Chain("ThreadLock system error : ") + Chain(strerror(errno));
	throw Exception(EXLOC, msg);
    }

#endif



}


void ThreadLock::writeLock()
{   

#ifdef HAVE_MINGW32
    EnterCriticalSection (& _rwlock);
#else

    NanoTimer *pTimer = 0;

    if ( _doStat )
    {
	pTimer = new NanoTimer();
	pTimer->start();

	pthread_rwlock_wrlock(&_statlock);
	_numLockTry++;
	_writeLockCount++;
	pthread_rwlock_unlock(&_statlock);

    }

    int ret = pthread_rwlock_wrlock(&_rwlock);

    if ( _doStat )
    {
	pTimer->stop();

	pthread_rwlock_wrlock(&_statlock);
	_numLockTry--;
	_writeDelay += pTimer->getSum();
	pthread_rwlock_unlock(&_statlock);

	delete pTimer;
	
    }

    if  ( ret != 0 )
    {
	Chain msg = Chain("ThreadLock system error : ") + Chain(strerror(errno));
	throw Exception(EXLOC, msg);
    }


#endif
}

void ThreadLock::readLock(int timeout)
{   

#ifdef HAVE_MINGW32
    EnterCriticalSection (& _rwlock);
#else

    NanoTimer *pTimer = 0;
    
    if ( _doStat )
    {
	pTimer = new NanoTimer();
	pTimer->start();
	
	pthread_rwlock_wrlock(&_statlock);
	_numLockTry++;
	_readLockCount++;
	pthread_rwlock_unlock(&_statlock);

    }

    struct timespec delay;

    delay.tv_sec = 0 ; 
    delay.tv_nsec = _lockDelay * DELAYFAC;
    
    bool isLocked = false;
    bool isError = false;
    int locktime=0;
    Chain msg;

    while ( locktime < timeout && isLocked == false && isError == false)
    {
	int ret;
	ret = pthread_rwlock_tryrdlock(&_rwlock);
	if ( ret == 0 )
	{
	    isLocked = true;
	}
	else if ( ret == EBUSY )
	{
	   
	    if ( nanosleep(&delay, NULL) == -1 )
	    {
		msg = Chain("ThreadLock system error : ") + Chain(strerror(errno));
		isError=true;
	    }

	    locktime += _lockDelay;
	}
	else
	{
	    isError = true;
	    msg = Chain("ThreadLock system error : ") + Chain(strerror(errno));
	}
    }

    if ( _doStat )
    {
	pTimer->stop();
	
	pthread_rwlock_wrlock(&_statlock);
	_numLockTry--;
	_readDelay += pTimer->getSum();
	pthread_rwlock_unlock(&_statlock);
	
	delete pTimer;
    }

    if ( isLocked )
	return;

    if ( isError == false )
	msg = Chain("Lock timeout exceeded for <") + _id + Chain("> after ") + Chain(locktime) + Chain(" msec");

    throw Exception(EXLOC, msg);
#endif

}

void ThreadLock::writeLock(int timeout)
{   

#ifdef HAVE_MINGW32
    EnterCriticalSection (& _rwlock);
#else

    NanoTimer *pTimer = 0;
    
    if ( _doStat )
    {
	pTimer = new NanoTimer();
	pTimer->start();
	
	pthread_rwlock_wrlock(&_statlock);
	_numLockTry++;
	_writeLockCount++;
	pthread_rwlock_unlock(&_statlock);

    }

    struct timespec delay;

    delay.tv_sec = 0; 
    delay.tv_nsec = _lockDelay * DELAYFAC; 

    bool isLocked = false;
    bool isError = false;
    int locktime=0;
    Chain msg;

    while ( locktime < timeout && isLocked == false && isError == false )
    {
	int ret;
	ret = pthread_rwlock_trywrlock(&_rwlock);

	if ( ret == 0 )
	{
	    isLocked = true;
	}
	else if ( ret == EBUSY )
	{
	    if ( nanosleep(&delay, NULL) == -1 )
	    {
		isError = true;
		msg = Chain("ThreadLock system error : ") + Chain(strerror(errno));
	    }

	    locktime += _lockDelay;
	}
	else
	{
	    isError = true;
	    msg = Chain("ThreadLock system error : ") + Chain(strerror(errno));
	}
    }

    if ( _doStat )
    {
	pTimer->stop();
	
	pthread_rwlock_wrlock(&_statlock);
	_numLockTry--;
	_writeDelay += pTimer->getSum();
	pthread_rwlock_unlock(&_statlock);
	
	delete pTimer;
    }

    if ( isLocked )
	return;

    if ( isError == false )
	msg = Chain("Lock timeout exceeded for <") + _id + Chain("> after ") + Chain(locktime) + Chain(" msec");

    throw Exception(EXLOC, msg);
#endif

   
}

void ThreadLock::unlock()
{   

#ifdef HAVE_MINGW32
    LeaveCriticalSection (& _rwlock);
#else
    if ( pthread_rwlock_unlock(&_rwlock) != 0 )
    {
	Chain msg = Chain("ThreadLock system error : ") + Chain(strerror(errno));
	throw Exception(EXLOC, msg);
    }
#endif	    
}


long ThreadLock::numReadLock()
{
    return _readLockCount;
}

long ThreadLock::numWriteLock()
{
    return _writeLockCount;
}

long ThreadLock::numLockTry()
{
    return _numLockTry;
}

long ThreadLock::sumReadDelay()
{
    return _readDelay;
}

long ThreadLock::sumWriteDelay()
{
    return _writeDelay;
}
